import numpy as np

from ffcv import DatasetWriter
from ffcv.fields import RGBImageField, JSONField, NDArrayField

from ffcv_pl.datasets.image import ImageDataset
from ffcv_pl.datasets.image_label import ImageLabelDataset
from ffcv_pl.datasets.image_array import ImageArrayDataset


def create_image_dataset(train_folder: str = None, validation_folder: str = None, test_folder: str = None):
    """
    create train/validation and optionally test .beton file to the father directory of each path.
    E.G.:
        train_folder = '/home/datasets/imagenet/train/' --> '/home/datasets/imagenet/train.beton'
    """

    folders = [train_folder, validation_folder, test_folder]
    names = ['train.beton', 'validation.beton', 'test.beton']

    for f, n in zip(folders, names):

        if f is not None:
            path = f"{'/'.join(f.split('/')[:-2])}/{n}"
            dataset = ImageDataset(f)

            # official guidelines: https://docs.ffcv.io/writing_datasets.html
            print(f'creating dataset from folder: {f}\tto file: {path}')
            writer = DatasetWriter(path, {'image': RGBImageField(write_mode='jpg')})
            writer.from_indexed_dataset(dataset)
        else:
            print(f'No folder name provided for {n}')


def create_image_label_dataset(train_folder: str = None, validation_folder: str = None, test_folder: str = None):
    """
    create train/validation and optionally test .beton file to the father directory of each path.
    E.G.:
        train_folder = '/home/datasets/imagenet/train/' --> '/home/datasets/imagenet/train.beton'

    folder must be organized as follows (to correctly get class name):
            /home/datasets/imagenet/train/
                                    class1/
                                            images
                                    class2/
                                            images
                                    classn/
                                            images
    """

    folders = [train_folder, validation_folder, test_folder]
    names = ['train.beton', 'validation.beton', 'test.beton']

    for f, n in zip(folders, names):

        if f is not None:
            path = f"{'/'.join(f.split('/')[:-2])}/{n}"
            dataset = ImageLabelDataset(f)

            # official guidelines: https://docs.ffcv.io/writing_datasets.html
            print(f'creating dataset from folder: {f}\tto file: {path}')
            writer = DatasetWriter(path, {'image': RGBImageField(write_mode='jpg'), 'label': JSONField()})
            writer.from_indexed_dataset(dataset)
        else:
            print(f'No folder name provided for {n}')


def create_image_array_dataset(array_shape: tuple, array_dtype: np.dtype = np.dtype('float32'),
                               train_folder: str = None, validation_folder: str = None, test_folder: str = None):
    """
    create train/validation and optionally test .beton file to the father directory of each path.
    E.G.:
        train_folder = '/home/datasets/imagenet/train/' --> '/home/datasets/imagenet/train.beton'

    folder must be organized as follows (to correctly get image/numpy array couples):
            /home/datasets/my_custom_dataset/train/
                                                /0001.npy
                                                /0001.png
                                                /0002.npy
                                                /0002.png
                                                ...
                                                /wxyz.npy
                                                /wxyz.png
    :param array_shape: shape of numpy arrays (all must have the same shape)
    :param array_dtype: dtype of numpy arrays (all must have the same dtype - default to np.dtype('float32'))
    """

    folders = [train_folder, validation_folder, test_folder]
    names = ['train.beton', 'validation.beton', 'test.beton']

    for f, n in zip(folders, names):

        if f is not None:
            path = f"{'/'.join(f.split('/')[:-2])}/{n}"
            dataset = ImageArrayDataset(f)

            # official guidelines: https://docs.ffcv.io/writing_datasets.html
            print(f'creating dataset from folder: {f}\tto file: {path}')
            writer = DatasetWriter(path, {'image': RGBImageField(write_mode='jpg'),
                                          'array': NDArrayField(array_dtype, array_shape)})
            writer.from_indexed_dataset(dataset)
        else:
            print(f'No folder name provided for {n}')



from ffcv.pipeline.operation import Operation
from ffcv.pipeline.allocation_query import AllocationQuery
from ffcv.pipeline.state import State
from dataclasses import replace

from typing import Tuple, Optional, Callable

import torch


class DivideImage255(Operation):

    def __init__(self, dtype: torch.dtype):
        """
        Divide input torch tensor by 255. Intended to get images in range [0, 1]
        :param dtype: type of returned images. Accepts torch.float16, torch.float32, torch.float64
        """
        super().__init__()

        assert dtype == torch.float16 or dtype == torch.float32 or dtype == torch.float64, \
            f'wrong dtype passed: {dtype}'

        self.dtype = dtype

    def generate_code(self) -> Callable:
        def divide(image, dst):
            dst = image.to(self.dtype) / 255.
            return dst

        divide.is_parallel = True

        return divide

    def declare_state_and_memory(self, previous_state: State) -> Tuple[State, Optional[AllocationQuery]]:

        return replace(previous_state, dtype=self.dtype), None



import torch
from torch.utils.data import Dataset
import pytorch_lightning as pl
from pathlib import Path
from PIL import Image
import numpy as np

from ffcv.loader import Loader, OrderOption
from ffcv.transforms import ToTensor, ToTorchImage
from ffcv.fields.decoders import CenterCropRGBImageDecoder, NDArrayDecoder

from ffcv_pl.ffcv_utils.custom_augmentations import DivideImage255


class ImageArrayDataset(Dataset):

    def __init__(self, folder: str):
        """
        create a dataset to return (image, np_array) pairs. May be used for example to associate GAN generations and
        the corresponding latent.

        Assumes that the folders contains couples of image files and .npy files, with the same name

        E.G.
        /home/dataset/gan_generations/
                                /0001.npy
                                /0001.png
                                /0002.npy
                                /0002.png
                                ...
                                /wxyz.npy
                                /wxyz.png

        :param folder: path to images in [.png, .jpg, .jPEG] formats and [.npy] array files.
        """
        self.image_names = [str(p.resolve()) for p in Path(folder).rglob('*.JPEG')] + \
                                  [str(p.resolve()) for p in Path(folder).rglob('*.jpg')] + \
                                  [str(p.resolve()) for p in Path(folder).rglob('*.png')]

        # check correctness
        npy_names = [str(p.resolve()) for p in Path(folder).rglob('*.npy')]

        assert len(self.image_names) == len(npy_names), \
            f'In "{folder}" there are {len(self.image_names)} images and {len(npy_names)} numpy arrays.\n' \
            f'Please ensure that each image has only one corresponding array and vice-versa.'

    def __len__(self):
        return len(self.image_names)

    def __getitem__(self, index):
        """
        :param index: the image index to return
        :return: the corresponding (image, npy) couple
        """

        # load image
        image_tensor = np.array(Image.open(self.image_names[index]).convert('RGB'))

        # load array
        np_array = np.load('.'.join(self.image_names[index].split('.')[:-1]) + '.npy')

        return image_tensor, np_array


class ImageArrayDataModule(pl.LightningDataModule):

    def __init__(self, train_file: str, val_file: str, test_file: str, image_size: int, dtype: torch.dtype,
                 batch_size: int, num_workers: int, is_dist: bool, seed: int):
        """
        :param train_file: path to .beton train file
        :param val_file: path to .beton validation file
        :param test_file: path to .beton test file
        :param image_size: all loaded images will be returned with this size
        :param dtype: torch desired type for returned images. May be one of torch.float16, torch.float32, torch.float64
        """

        super().__init__()

        self.train_file = train_file
        self.val_file = val_file
        self.test_file = test_file

        self.image_size = image_size
        self.batch_size = batch_size
        self.num_workers = num_workers
        self.seed = seed
        self.is_dist = is_dist
        self.dtype = dtype

        self.pipeline = None

    def setup(self, stage=None):
        # prepare data
        img_pipeline = [CenterCropRGBImageDecoder(output_size=(self.image_size, self.image_size), ratio=1.),
                        ToTensor(),
                        ToTorchImage(),
                        DivideImage255(self.dtype),
                        ]

        array_pipeline = [NDArrayDecoder()]

        self.pipeline = {
            'image': img_pipeline,
            'array': array_pipeline
        }

    def train_dataloader(self):
        return Loader(self.train_file, batch_size=self.batch_size, num_workers=self.num_workers,
                      order=OrderOption.RANDOM, pipelines=self.pipeline, distributed=self.is_dist, seed=self.seed)

    def val_dataloader(self):
        return Loader(self.val_file, batch_size=self.batch_size, num_workers=self.num_workers,
                      order=OrderOption.SEQUENTIAL, pipelines=self.pipeline, distributed=self.is_dist, seed=self.seed)

    def test_dataloader(self):
        return Loader(self.test_file, batch_size=self.batch_size, num_workers=self.num_workers,
                      order=OrderOption.SEQUENTIAL, pipelines=self.pipeline, distributed=self.is_dist, seed=self.seed)



from setuptools import setup
from pathlib import Path

# set readme
this_directory = Path(__file__).parent
long_description = (this_directory / "../readme.md").read_text()

setup(name='ffcv_pl',
      version='0.1.5',
      description='manage fast data loading with ffcv and pytorch lightning',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/SerezD/ffcv_pytorch_lightning',
      author='DSerez',
      license='MIT',
      zip_safe=False)
